# cython: language_level=3

def set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic):
    cpu.pc = pc
    cpu.a = a
    cpu.x = x
    cpu.y = y
    cpu.sp = sp
    cpu.n = n
    cpu.z = z
    cpu.c = c
    cpu.v = v
    cpu.instruction_count = ic


def run(cpu):
    cdef unsigned char* rom = cpu.rom
    cdef unsigned char* ram = cpu.ram
    cdef unsigned short pc = cpu.pc
    cdef unsigned char a = cpu.a
    cdef unsigned char x = cpu.x
    cdef unsigned char y = cpu.y
    cdef unsigned char sp = cpu.sp
    cdef unsigned char n = cpu.n
    cdef unsigned char z = cpu.z
    cdef unsigned char c = cpu.c
    cdef unsigned char v = cpu.v
    cdef unsigned int ic = cpu.instruction_count

    cdef unsigned char opc
    cdef unsigned char addressing
    cdef unsigned char operand
    cdef unsigned short address
    cdef unsigned char value

    # Instruction loop
    while 1:
        # Read opcode
        ic += 1
        opc = rom[pc - 0x8000]
        pc += 1

        # No operand

        # Increment/decrement

        # INX
        if opc == 0xe8:
            x += 1
            n = x >> 7
            z = x == 0
            continue
        # DEX
        elif opc == 0xca:
            x -= 1
            n = x >> 7
            z = x == 0
            continue
        # INY
        elif opc == 0xc8:
            y += 1
            n = y >> 7
            z = y == 0
            continue
        # DEY
        elif opc == 0x88:
            y -= 1
            n = y >> 7
            z = y == 0
            continue

        # Shifting

        # ASL ACC
        elif opc == 0x0a:
            c = a >> 7
            a <<= 1
            n = a >> 7
            z = a == 0
            continue
        # LSR ACC
        elif opc == 0x4a:
            c = a & 0x01
            a >>= 1
            n = a >> 7
            z = a == 0
            continue
        # ROL ACC
        elif opc == 0x2a:
            value = c
            c = a >> 7
            a <<= 1
            a |= value
            n = a >> 7
            z = a == 0
            continue
        # ROR ACC
        elif opc == 0x6a:
            value = c
            c = a & 0x01
            a >>= 1
            a |= value << 7
            n = a >> 7
            z = a == 0
            continue

        # Flow control

        # NOP
        elif opc == 0xea:
            continue
        # RTI/BRK
        elif opc in (0x00, 0x40):
            break
        # RTS
        elif opc == 0x60:
            sp += 1
            pc = ram[0x0100 | sp]
            sp += 1
            pc |= ram[0x0100 | sp] << 8
            pc += 1
            continue
        # PHA
        elif opc == 0x48:
            ram[0x0100 | sp] = a
            sp -= 1
            continue
        # PLA
        elif opc == 0x68:
            sp += 1
            a = ram[0x0100 | sp]
            n = a >> 7
            z = a == 0
            continue
        # PHP
        elif opc == 0x08:
            value = (n << 7) | (v << 6) | (z << 1) | (c << 0)
            ram[0x0100 | sp] = value
            sp -= 1
            continue
        # PLP
        elif opc == 0x28:
            sp += 1
            value = ram[0x0100 | sp]
            n = (value & 0x80) != 0
            v = (value & 0x40) != 0
            z = (value & 0x02) != 0
            c = (value & 0x01) != 0
            continue

        # Flags

        # CLC
        elif opc == 0x18:
            c = 0
            continue
        # SEC
        elif opc == 0x38:
            c = 1
            continue
        # CLI/SEI/CLV/CLD/SED
        elif opc in (0x58, 0x78, 0xB8, 0xD8, 0xF8):
            continue
        # TAX
        elif opc == 0xaa:
            x = a
            n = a >> 7
            z = a == 0
            continue

        # Transfer

        # TXA
        elif opc == 0x8a:
            a = x
            n = a >> 7
            z = a == 0
            continue
        # TAY
        elif opc == 0xa8:
            y = a
            n = a >> 7
            z = a == 0
            continue
        # TYA
        elif opc == 0x98:
            a = y
            n = a >> 7
            z = a == 0
            continue
        # TXS
        elif opc == 0x9a:
            sp = x
            continue
        # TSX
        elif opc == 0xba:
            x = sp
            n = x >> 7
            z = x == 0
            continue

        # Byte operand
        operand = rom[pc - 0x8000]
        pc += 1

        addressing = (opc & 0b00011100) >> 2
        # ZPY
        if opc in (0x96, 0xb6):
            operand += y
            address = operand
            value = ram[address]
        # ZPX
        elif addressing == 0x05:
            operand += x
            address = operand
            value = ram[address]
        # ZPG
        elif addressing == 0x01:
            address = operand
            value = ram[address]
        # INX
        elif (opc & 0x0f) == 0x01 and addressing == 0x00:
            operand += x
            address = ram[operand]
            operand += 1
            address |= ram[operand] << 8
            if opc == 0x81:
                value = 0
            elif address < 0x800:
                value = ram[address]
            elif address > 0x8000:
                value = rom[address - 0x8000]
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                value = cpu.cpu_read(address)
        # INY
        elif (opc & 0x0f) == 0x01 and addressing == 0x04:
            address = ram[operand]
            operand += 1
            address |= ram[operand] << 8
            address += y
            if opc == 0x91:
                value = 0
            elif address < 0x800:
                value = ram[address]
            elif address > 0x8000:
                value = rom[address - 0x8000]
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                value = cpu.cpu_read(address)
        # IMM / REL
        else:
            address = 0
            value = operand

        # Branching

        # BPL
        if opc == 0x10:
            if not n:
                pc += <char>value
            continue
        # BMI
        elif opc == 0x30:
            if n:
                pc += <char>value
            continue
        # BNE
        elif opc == 0xd0:
            if not z:
                pc += <char>value
            continue
        # BEQ
        elif opc == 0xf0:
            if z:
                pc += <char>value
            continue
        # BCC
        elif opc == 0x90:
            if not c:
                pc += <char>value
            continue
        # BCS
        elif opc == 0xB0:
            if c:
                pc += <char>value
            continue
        # BVC
        elif opc == 0x50:
            if not v:
                pc += <char>value
            continue
        # BVS
        elif opc == 0x70:
            if v:
                pc += <char>value
            continue

        # Store

        # STA ZPG/ZPX
        elif opc in (0x85, 0x95):
            ram[address] = a
            continue
        # STX ZPG
        elif opc in (0x86, 0x96):
            ram[address] = x
            continue
        # STY ZPG
        elif opc in (0x84, 0x94):
            ram[address] = y
            continue
        # STA INX/INY
        elif opc in (0x81, 0x91):
            if address < 0x800:
                ram[address] = a
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, a)
            continue

        # Load

        # LDA IMM/ZPG/ZPX/INX/INY
        elif opc in (0xa9, 0xa5, 0xb5, 0xa1, 0xb1):
            a = value
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # LDX IMM/ZPG/ZPY
        elif opc in (0xa2, 0xa6, 0xb6):
            x = value
            n = (x & 0x80) != 0
            z = x == 0
            continue
        # LDY IMM/ZPG/ZPX
        elif opc in (0xa0, 0xa4, 0xb4):
            y = value
            n = (y & 0x80) != 0
            z = y == 0
            continue

        # Bitwise operation

        # ORA IMM/ZPG/ZPX/INX/INY
        elif opc in (0x09, 0x05, 0x15, 0x01, 0x11):
            a |= value
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # AND IMM/ZPG/ZPX/INX/INY
        elif opc in (0x29, 0x25, 0x35, 0x21, 0x31):
            a &= value
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # EOR IMM/ZPG/ZPX/INX/INY
        elif opc in (0x49, 0x45, 0x55, 0x41, 0x51):
            a ^= value
            n = (a & 0x80) != 0
            z = a == 0
            continue

        # Arithmetic operation

        # ADC/SBC IMM/ZPG/ZPX/INX/INY
        elif opc in (0x69, 0x65, 0x75, 0x61, 0x71, 0xe9, 0xe5, 0xf5, 0xe1, 0xf1):
            # Invert for SBC
            if opc in (0xe9, 0xe5, 0xf5, 0xe1, 0xf1):
                value = ~value
            # Save signs of operand
            n = a >> 8
            z = value >> 8
            # Add carry
            a += c
            c = a < c
            # Add value
            a += value
            c |= a < value
            # Compute overflow
            v = a >> 8
            v = (v ^ n) & (v ^ z)
            # Set N and Z
            n = (a & 0x80) != 0
            z = a == 0
            continue


        # Increment / decrement

        # INC ZPG/ZPX
        elif opc in (0xe6, 0xf6):
            value += 1
            n = (value & 0x80) != 0
            z = value == 0
            ram[address] = value
            continue
        # DEC ZPG/ZPX
        elif opc in (0xc6, 0xd6):
            value -= 1
            n = (value & 0x80) != 0
            z = value == 0
            ram[address] = value
            continue

        # Shifting

        # ASL ZPG
        elif opc in (0x06, 0x16):
            c = value >> 7
            value <<= 1
            n = value >> 7
            z = value == 0
            ram[address] = value
            continue
        # LSR ZPG
        elif opc in (0x46, 0x56):
            c = value & 0x01
            value >>= 1
            n = value >> 7
            z = value == 0
            ram[address] = value
            continue
        # ROL ZPG
        elif opc in (0x26, 0x36):
            n = c
            c = value >> 7
            value <<= 1
            value |= n
            n = value >> 7
            z = value == 0
            ram[address] = value
            continue
        # ROR ZPG
        elif opc in (0x66, 0x76):
            n = c
            c = value & 0x01
            value >>= 1
            value |= n << 7
            n = value >> 7
            z = value == 0
            ram[address] = value
            continue

        # Comparison

        # CMP IMM/ZPG/ZPX
        elif opc in (0xc9, 0xc5, 0xd5):
            c = a >= value
            value = a - value
            n = (value & 0x80) != 0
            z = value == 0
            continue
        # CPX IMM/ZPG
        elif opc in (0xe0, 0xe4):
            c = x >= value
            value = x - value
            n = (value & 0x80) != 0
            z = value == 0
            continue
        # CPY IMM/ZPG
        elif opc in (0xc0, 0xc4):
            c = y >= value
            value = y - value
            n = (value & 0x80) != 0
            z = value == 0
            continue
        # BIT ZPG
        elif opc == 0x24:
            z = (a & value) == 0
            n = (value & 0x80) != 0
            v = (value & 0x40) != 0
            continue

        # Word operand
        address = (rom[pc - 0x8000] << 8) | operand
        pc += 1

        # ABX and ABY
        if addressing == 0x06 or opc == 0xbe:
            address += y
        elif addressing == 0x07:
            address += x

        # Store

        # STA ABS/ABX/ABY
        if opc in (0x8d, 0x9d, 0x99):
            if address < 0x800:
                ram[address] = a
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, a)
            continue
        # STX ABS
        if opc == 0x8e:
            if address < 0x800:
                ram[address] = x
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, x)
            continue
        # STY ABS
        if opc == 0x8c:
            if address < 0x800:
                ram[address] = y
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, y)
            continue

        # Flow control

        # JSR
        elif opc == 0x20:
            pc -= 1
            ram[0x0100 | sp] = pc >> 8
            sp -= 1
            ram[0x0100 | sp] = pc & 0xff
            sp -= 1
            pc = address
            continue
        # JMP
        elif opc == 0x4c:
            if pc == address + 3:
                pc = address
                break
            pc = address
            continue
        # JMP IND
        elif opc == 0x6c:
            if address < 0x800:
                value = ram[address]
            elif address >= 0x8000:
                value = rom[address - 0x8000]
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                value = cpu.cpu_read(address)
            address += 1
            if address < 0x800:
                pc = (ram[address] << 8) | value
            elif address >= 0x8000:
                pc = (rom[address - 0x8000] << 8) | value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                pc = (rom[cpu.cpu_read(address)] << 8) | value
            continue

        # Get value at absolute address
        if address < 0x800:
            value = ram[address]
        elif address >= 0x8000:
            value = rom[address - 0x8000]
        else:
            set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
            value = cpu.cpu_read(address)

        # Load

        # LDA ABS/ABX/ABY
        if opc in (0xad, 0xbd, 0xb9):
            a = value
            n = value >> 7
            z = value == 0
            continue
        # LDX ABS/ABY
        elif opc in (0xae, 0xbe):
            x = value
            n = value >> 7
            z = value == 0
            continue
        # LDY ABS/ABX
        elif opc in (0xac, 0xbc):
            y = value
            n = value >> 7
            z = value == 0
            continue
        # CMP ABS/ABX/ABY
        elif opc in (0xcd, 0xdd, 0xd9):
            c = a >= value
            value = a - value
            n = (value & 0x80) != 0
            z = value == 0
            continue
        # CPX ABS
        elif opc == 0xec:
            c = x >= value
            value = x - value
            n = (value & 0x80) != 0
            z = value == 0
            continue
        # CPY ABS
        elif opc == 0xcc:
            c = y >= value
            value = y - value
            n = (value & 0x80) != 0
            z = value == 0
            continue
        # INC ABS/ABX
        elif opc in (0xee, 0xfe):
            value += 1
            n = (value & 0x80) != 0
            z = value == 0
            if address < 0x800:
                ram[address] = value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, value)
            continue
        # DEC ABS/ABX
        elif opc in (0xce, 0xde):
            value -= 1
            n = (value & 0x80) != 0
            z = value == 0
            if address < 0x800:
                ram[address] = value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, value)
            continue
        # ADC/SBC ABS/ABX/ABY
        elif opc in (0x6d, 0x7d, 0x79, 0xed, 0xfd, 0xf9):
            # Invert for SBC
            if opc in ((0xed, 0xfd, 0xf9)):
                value = ~value
            # Save signs of operand
            n = a >> 8
            z = value >> 8
            # Add carry
            a += c
            c = a < c
            # Add value
            a += value
            c |= a < value
            # Compute overflow
            v = a >> 8
            v = (v ^ n) & (v ^ z)
            # Set N and Z
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # ORA ABS/ABX/ABY
        elif opc in (0x0d, 0x1d, 0x19):
            a |= value
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # AND ABS/ABX/ABY
        elif opc in (0x2d, 0x3d, 0x39):
            a &= value
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # EOR ABS/ABX/ABY
        elif opc in (0x4d, 0x5d, 0x59):
            a ^= value
            n = (a & 0x80) != 0
            z = a == 0
            continue
        # ASL ABS
        elif opc in (0x0e, 0x1e):
            c = value >> 7
            value <<= 1
            n = value >> 7
            z = value == 0
            if address < 0x800:
                ram[address] = value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, value)
            continue
        # LSR ABS
        elif opc in (0x4e, 0x5e):
            c = value & 0x01
            value >>= 1
            n = value >> 7
            z = value == 0
            if address < 0x800:
                ram[address] = value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, value)
            continue
        # ROL ABS
        elif opc in (0x2e, 0x3e):
            n = c
            c = value >> 7
            value <<= 1
            value |= n
            n = value >> 7
            z = value == 0
            if address < 0x800:
                ram[address] = value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, value)
            continue
        # ROR ABS
        elif opc in (0x6e, 0x7e):
            n = c
            c = value & 0x01
            value >>= 1
            value |= n << 7
            n = value >> 7
            z = value == 0
            if address < 0x800:
                ram[address] = value
            else:
                set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)
                cpu.cpu_write(address, value)
            continue
        # BIT ABS
        elif opc == 0x2c:
            z = (a & value) == 0
            n = (value & 0x80) != 0
            v = (value & 0x40) != 0
            continue

        # Opcode not supported
        pc -= 3
        ic -= 1
        break

    # Set the value back to the CPU instance
    set_cpu_attributes(cpu, pc, a, x, y, sp, n, z, c, v, ic)

    # Except RTI or JMP
    if opc not in (0x40, 0x4c):
        raise ValueError(f"Invalid opcode: 0x{opc:02x}")
    return opc



# cython: language_level=3

cimport numpy as np
from numpy import pi


cdef float[31] pulse_table = [
    0.0,
    0.011609139523578026,
    0.022939481268011527,
    0.03400094921689606,
    0.04480300187617261,
    0.05535465924895688,
    0.06566452795600367,
    0.07574082464884459,
    0.08559139784946236,
    0.09522374833850243,
    0.10464504820333041,
    0.11386215864759427,
    0.12288164665523155,
    0.13170980059397538,
    0.14035264483627205,
    0.1488159534690486,
    0.15710526315789472,
    0.16522588522588522,
    0.1731829170024174,
    0.18098125249301955,
    0.18862559241706162,
    0.19612045365662886,
    0.20347017815646784,
    0.21067894131185272,
    0.21775075987841944,
    0.2246894994354535,
    0.2314988814317673,
    0.23818248984115256,
    0.2447437774524158,
    0.2511860718171926,
    0.25751258087706685,
]

cdef float[202] tnd_table = [
    0.0,
    0.006699823979696262,
    0.01334502018019487,
    0.01993625400950099,
    0.026474180112418616,
    0.032959442587297105,
    0.03939267519756107,
    0.04577450157816932,
    0.05210553543714433,
    0.05838638075230885,
    0.06461763196336215,
    0.07079987415942428,
    0.07693368326217241,
    0.08301962620468999,
    0.08905826110614481,
    0.09505013744240969,
    0.10099579621273477,
    0.10689577010257789,
    0.11275058364269584,
    0.11856075336459644,
    0.12432678795244785,
    0.1300491883915396,
    0.13572844811338536,
    0.1413650531375568,
    0.1469594822103333,
    0.15251220694025122,
    0.15802369193063237,
    0.16349439490917161,
    0.16892476685465738,
    0.1743152521209005,
    0.1796662885579421,
    0.18497830763060993,
    0.19025173453449087,
    0.19548698830938505,
    0.20068448195030472,
    0.20584462251608032,
    0.2109678112356332,
    0.2160544436119733,
    0.2211049095239788,
    0.22611959332601225,
    0.2310988739454269,
    0.23604312497801538,
    0.24095271478145042,
    0.24582800656676793,
    0.25066935848793903,
    0.25547712372957787,
    0.2602516505928307,
    0.26499328257948945,
    0.26970235847437257,
    0.27437921242601526,
    0.27902417402570834,
    0.28363756838492643,
    0.2882197162111822,
    0.292770933882345,
    0.29729153351945914,
    0.3017818230580978,
    0.3062421063182866,
    0.31067268307302937,
    0.31507384911547015,
    0.3194458963247213,
    0.32378911273039,
    0.3281037825758322,
    0.3323901863801631,
    0.33664860099905314,
    0.3408792996843372,
    0.34508255214246325,
    0.349258624591807,
    0.3534077798188791,
    0.3575302772334479,
    0.36162637292260397,
    0.3656963197037888,
    0.3697403671768112,
    0.3737587617748739,
    0.37775174681463214,
    0.38171956254530554,
    0.38566244619686446,
    0.3895806320273106,
    0.3934743513690717,
    0.3973438326745308,
    0.40118930156070615,
    0.405010980853104,
    0.4088090906287582,
    0.41258384825847705,
    0.4163354684483128,
    0.42006416328027124,
    0.4237701422522769,
    0.42745361231741014,
    0.4311147779224318,
    0.4347538410456096,
    0.43837100123386197,
    0.4419664556392331,
    0.44554039905471293,
    0.44909302394941686,
    0.4526245205031371,
    0.45613507664027986,
    0.4596248780632002,
    0.4630941082849479,
    0.4665429486614358,
    0.46997157842304194,
    0.47338017470565896,
    0.4767689125811996,
    0.48013796508757145,
    0.48348750325813084,
    0.48681769615062515,
    0.49012871087563703,
    0.493420712624537,
    0.49669386469695664,
    0.49994832852779125,
    0.5031842637137408,
    0.5064018280393993,
    0.5096011775029012,
    0.5127824663411329,
    0.5159458470545188,
    0.5190914704313901,
    0.5222194855719443,
    0.5253300399118033,
    0.528423279245178,
    0.5314993477476477,
    0.5345583879985607,
    0.5376005410030638,
    0.5406259462137686,
    0.5436347415520602,
    0.5466270634290563,
    0.5496030467662235,
    0.5525628250156552,
    0.5555065301800212,
    0.5584342928321915,
    0.5613462421345432,
    0.5642425058579547,
    0.5671232104004943,
    0.5699884808058077,
    0.5728384407812124,
    0.5756732127155,
    0.5784929176964575,
    0.5812976755281083,
    0.5840876047476803,
    0.5868628226423054,
    0.5896234452654553,
    0.5923695874531196,
    0.595101362839729,
    0.5978188838738291,
    0.6005222618335111,
    0.6032116068415997,
    0.6058870278806079,
    0.6085486328074569,
    0.6111965283679723,
    0.6138308202111536,
    0.6164516129032258,
    0.6190590099414757,
    0.6216531137678758,
    0.6242340257825014,
    0.6268018463567424,
    0.6293566748463153,
    0.6318986096040777,
    0.6344277479926501,
    0.6369441863968464,
    0.6394480202359187,
    0.6419393439756177,
    0.6444182511400732,
    0.6468848343234979,
    0.6493391852017159,
    0.6517813945435207,
    0.6542115522218658,
    0.6566297472248885,
    0.659036067666773,
    0.6614306007984521,
    0.6638134330181533,
    0.6661846498817908,
    0.6685443361132047,
    0.670892575614252,
    0.6732294514747513,
    0.6755550459822829,
    0.6778694406318475,
    0.6801727161353863,
    0.6824649524311629,
    0.684746228693012,
    0.6870166233394548,
    0.6892762140426848,
    0.6915250777374256,
    0.693763290629662,
    0.6959909282052493,
    0.6982080652383982,
    0.7004147758000423,
    0.7026111332660865,
    0.70479721032554,
    0.7069730789885358,
    0.7091388105942369,
    0.7112944758186339,
    0.7134401446822323,
    0.7155758865576349,
    0.7177017701770176,
    0.7198178636395035,
    0.7219242344184336,
    0.7240209493685391,
    0.7261080747330146,
    0.7281856761504939,
    0.7302538186619317,
    0.7323125667173908,
    0.734361984182737,
    0.7364021343462434,
    0.7384330799251054,
    0.7404548830718675,
 ]



def apu_mixer(
    apu,
    np.ndarray[np.uint8_t, ndim=1] pulse1,
    np.ndarray[np.uint8_t, ndim=1] pulse2,
    np.ndarray[np.uint8_t, ndim=1] triangle,
    np.ndarray[np.uint8_t, ndim=1] noise,
    np.ndarray[np.uint8_t, ndim=1] dmc,
    np.ndarray[np.int16_t, ndim=2] output,
):
    cdef unsigned int ticks_in_frame = pulse1.shape[0]
    cdef unsigned int i
    cdef short value

    cdef float[14890] mixer_out
    cdef float current_in, current_out, alpha, omega
    cdef float dt = 1/(60 * ticks_in_frame)

    cdef float filter1_previous_in = apu.filter1_previous_in
    cdef float filter1_previous_out = apu.filter1_previous_out
    cdef float filter2_previous_in = apu.filter2_previous_in
    cdef float filter2_previous_out = apu.filter2_previous_out
    cdef float filter3_previous_in = apu.filter3_previous_in
    cdef float filter3_previous_out = apu.filter3_previous_out

    # Mixing
    for i in range(ticks_in_frame):
        mixer_out[i] = (
            pulse_table[pulse1[i] + pulse2[i]] +
            tnd_table[3 * triangle[i] + 2 * noise[i] + dmc[i]]
        )

    # High pass filter 1 (90 Hz)
    if apu.filter1_enabled:
        omega = 2.0 * pi * apu.filter1_cutoff
        alpha = 1 / (1 + omega * dt)
        for i in range(1, ticks_in_frame):
            current_in = mixer_out[i]
            current_out = (filter1_previous_out + current_in - filter1_previous_in) * alpha
            mixer_out[i] = current_out
            filter1_previous_in = current_in
            filter1_previous_out = current_out

    # High pass filter 2 (442 Hz)
    if apu.filter2_enabled:
        omega = 2.0 * pi * apu.filter2_cutoff
        alpha = 1 / (1 + omega * dt)
        for i in range(1, ticks_in_frame):
            current_in = mixer_out[i]
            current_out = (filter2_previous_out + current_in - filter2_previous_in) * alpha
            mixer_out[i] = current_out
            filter2_previous_in = current_in
            filter2_previous_out = current_out

    # Low pass filter 3 (14 kHz)
    if apu.filter3_enabled:
        omega = 2.0 * pi * apu.filter3_cutoff
        alpha = (omega * dt) / (1 + omega * dt)
        for i in range(1, ticks_in_frame):
            current_in = mixer_out[i]
            current_out = filter3_previous_out + (current_in - filter3_previous_out) * alpha
            mixer_out[i] = current_out
            filter3_previous_in = current_in
            filter3_previous_out = current_out

    for i in range(ticks_in_frame):
        value = <short>(mixer_out[i] * 32768)
        output[(i<<1)|0,0] = value
        output[(i<<1)|0,1] = value
        output[(i<<1)|1,0] = value
        output[(i<<1)|1,1] = value

    apu.filter1_previous_in = filter1_previous_in
    apu.filter1_previous_out = filter1_previous_out
    apu.filter2_previous_in = filter2_previous_in
    apu.filter2_previous_out = filter2_previous_out
    apu.filter3_previous_in = filter3_previous_in
    apu.filter3_previous_out = filter3_previous_out


def generate_pulse(
    pulse,
    np.ndarray[np.uint8_t, ndim=1] pulse_out,
):
    cdef unsigned int ticks_in_frame = len(pulse_out)
    cdef unsigned int i, value, current_envelope

    # Parameters
    cdef unsigned short duty = pulse.duty
    cdef unsigned short volume = pulse.volume
    cdef unsigned short load_timer = pulse.load_timer
    cdef unsigned short sweep_negate = pulse.sweep_negate
    cdef unsigned short sweep_period = pulse.sweep_period
    cdef unsigned short sweep_enabled = pulse.sweep_enabled
    cdef unsigned short constant_volume = pulse.constant_volume
    cdef unsigned short sweep_reload_flag = pulse.sweep_reload_flag
    cdef unsigned short sweep_shift_count = pulse.sweep_shift_count
    cdef unsigned short load_length_counter = pulse.load_length_counter
    cdef unsigned short length_counter_halt = pulse.length_counter_halt

    # Internal state
    cdef unsigned short enabled = pulse.enabled
    cdef unsigned short start_flag = pulse.start_flag
    cdef unsigned short current_tick = pulse.current_tick
    cdef unsigned short current_timer = pulse.current_timer
    cdef unsigned short sweep_divider = pulse.sweep_divider
    cdef unsigned short length_counter = pulse.length_counter
    cdef unsigned short divider_period = pulse.divider_period
    cdef unsigned short current_sequencer = pulse.current_sequencer
    cdef unsigned short decay_level_counter = pulse.decay_level_counter
    cdef unsigned short current_timer_period = pulse.current_timer_period

    cdef unsigned char[8] sequencer_table

    if duty == 0:
        sequencer_table = [0, 0, 0, 0, 0, 0, 0, 1]
    elif duty == 1:
        sequencer_table = [0, 0, 0, 0, 0, 0, 1, 1]
    elif duty == 2:
        sequencer_table = [0, 0, 0, 0, 1, 1, 1, 1]
    elif duty == 3:
        sequencer_table = [1, 1, 1, 1, 1, 1, 0, 0]
    else:
        assert False

    current_envelope = volume if constant_volume else decay_level_counter
    value = sequencer_table[current_sequencer] * current_envelope
    for i in range(ticks_in_frame):
        # Manage envelope
        if current_tick in (3728, 7456, 11185, 18640):
            if start_flag:
                start_flag = 0
                decay_level_counter = 15
                divider_period = volume
            elif divider_period != 0:
                divider_period -= 1
            else:
                divider_period = volume
                if decay_level_counter != 0:
                    decay_level_counter -= 1
                elif length_counter_halt == 1:
                    decay_level_counter = 15
            current_envelope = volume if constant_volume else decay_level_counter
        # Manage length counter
        if current_tick in (7456, 18640):
            if length_counter_halt == 0 and length_counter != 0:
                length_counter -= 1
            # Perform the sweep
            if sweep_divider == 0 and sweep_enabled and current_timer_period >= 8:
                if sweep_negate:
                    current_timer_period -= (current_timer_period >> sweep_shift_count)
                else:
                    current_timer_period += (current_timer_period >> sweep_shift_count)
            # Update sweep divider
            if sweep_divider == 0 or sweep_reload_flag:
                sweep_reload_flag = 0
                sweep_divider = sweep_period
            else:
                sweep_divider -= 1
        # Manage frame counter
        current_tick += 1
        if current_tick == 18641:
            current_tick = 0
        # Manage timer
        if current_timer != 0:
            current_timer -= 1
        else:
            current_timer = current_timer_period
            current_sequencer = 7 if current_sequencer == 0 else current_sequencer - 1
            value = sequencer_table[current_sequencer] * current_envelope
        # Manage length counter
        if enabled and length_counter > 0 and current_timer_period >= 8:
            pulse_out[i] = value
        else:
            pulse_out[i] = 0

    pulse.start_flag = start_flag
    pulse.current_tick = current_tick
    pulse.current_timer = current_timer
    pulse.sweep_divider = sweep_divider
    pulse.length_counter = length_counter
    pulse.divider_period = divider_period
    pulse.sweep_reload_flag = sweep_reload_flag
    pulse.current_sequencer = current_sequencer
    pulse.decay_level_counter = decay_level_counter
    pulse.current_timer_period = current_timer_period

def generate_triangle(
    triangle,
    np.ndarray[np.uint8_t, ndim=1] triangle_out,
):
    cdef unsigned int ticks_in_frame = len(triangle_out)
    cdef unsigned int i, value

    # Parameters
    cdef unsigned char enabled = triangle.enabled
    cdef unsigned short load_timer = triangle.load_timer
    cdef unsigned short load_counter = triangle.load_counter
    cdef unsigned short load_length_counter = triangle.load_length_counter
    cdef unsigned short length_counter_halt = triangle.length_counter_halt

    # Internal state
    cdef unsigned short current_tick = triangle.current_tick
    cdef unsigned short current_value = triangle.current_value
    cdef unsigned short current_timer = triangle.current_timer
    cdef unsigned short length_counter = triangle.length_counter
    cdef unsigned short current_counter = triangle.current_counter
    cdef unsigned short current_sequencer = triangle.current_sequencer
    cdef unsigned short counter_reload_flag = triangle.counter_reload_flag

    cdef unsigned char[32] sequencer_table = [
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
        15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
    ]

    for i in range(ticks_in_frame):
        # Manage volume
        if current_tick in (3728, 7456, 11185, 18640):
            if counter_reload_flag:
                counter_reload_flag = 0
                current_counter = load_counter
            elif current_counter != 0:
                current_counter -= 1
        # Manage length counter
        if current_tick in (7456, 18640):
            if length_counter_halt == 0 and length_counter != 0:
                length_counter -= 1
        # Manage frame counter
        current_tick += 1
        if current_tick == 18641:
            current_tick = 0
        # Manage timer
        if current_timer > 1:
            current_timer -= 2
        else:
            current_timer = load_timer if current_timer == 1 else load_timer - 1
            if current_counter != 0 and load_timer >= 2 and enabled and length_counter != 0:
                current_sequencer = 31 if current_sequencer == 0 else current_sequencer - 1
            current_value = sequencer_table[current_sequencer]
        triangle_out[i] = current_value

    triangle.current_value = current_value
    triangle.current_tick = current_tick
    triangle.current_timer = current_timer
    triangle.length_counter = length_counter
    triangle.current_counter = current_counter
    triangle.current_sequencer = current_sequencer
    triangle.counter_reload_flag = counter_reload_flag


def generate_noise(
    noise,
    np.ndarray[np.uint8_t, ndim=1] noise_out,
):
    cdef unsigned int ticks_in_frame = len(noise_out)
    cdef unsigned int i, current_envelope, shift, feedback

    # Parameters
    cdef unsigned short volume = noise.volume
    cdef unsigned short enabled = noise.enabled
    cdef unsigned short noise_mode = noise.noise_mode
    cdef unsigned short noise_period = noise.noise_period
    cdef unsigned short constant_volume = noise.constant_volume
    cdef unsigned short load_length_counter = noise.load_length_counter
    cdef unsigned short length_counter_halt = noise.length_counter_halt

    # Internal state
    cdef unsigned short start_flag = noise.start_flag
    cdef unsigned short current_tick = noise.current_tick
    cdef unsigned short current_timer = noise.current_timer
    cdef unsigned short length_counter = noise.length_counter
    cdef unsigned short shift_register = noise.shift_register
    cdef unsigned short divider_period = noise.divider_period
    cdef unsigned short decay_level_counter = noise.decay_level_counter

    shift = 6 if noise_mode else 1
    current_envelope = volume if constant_volume else decay_level_counter
    for i in range(ticks_in_frame):
        # Manage envelope
        if current_tick in (3728, 7456, 11185, 18640):
            if start_flag:
                start_flag = 0
                decay_level_counter = 15
                divider_period = volume
            elif divider_period != 0:
                divider_period -= 1
            else:
                divider_period = volume
                if decay_level_counter != 0:
                    decay_level_counter -= 1
                elif length_counter_halt == 1:
                    decay_level_counter = 15
            current_envelope = volume if constant_volume else decay_level_counter
        # Manage length counter
        if current_tick in (7456, 18640):
            if length_counter_halt == 0 and length_counter != 0:
                length_counter -= 1
        # Manage frame counter
        current_tick += 1
        if current_tick == 18641:
            current_tick = 0
        # Manage timer
        if current_timer != 0:
            current_timer -= 1
        else:
            current_timer = noise_period
            feedback = (shift_register ^ (shift_register >> shift)) & 0x01
            shift_register >>= 1
            shift_register |= feedback << 14
        # Manage length counter
        if enabled and (shift_register & 0x01) == 0 and length_counter != 0:
            noise_out[i] = current_envelope
        else:
            noise_out[i] = 0

    noise.start_flag = start_flag
    noise.current_tick = current_tick
    noise.current_timer = current_timer
    noise.length_counter = length_counter
    noise.shift_register = shift_register
    noise.divider_period = divider_period
    noise.decay_level_counter = decay_level_counter


from __future__ import annotations
from argparse import ArgumentParser, Namespace

import pickle
from copy import deepcopy
from enum import IntEnum
from functools import lru_cache
from dataclasses import dataclass, field
import zlib


import numpy as np
import numpy.typing as npt
from gambaterm.console import Console
from gambaterm.main import main as gambaterm_main

from . import nescpu
from . import nesppu
from . import nesapu


class InfiniteLoop(Exception):
    pass


@dataclass
class Cartridge:
    mapper: int
    mirroring: str
    cartridge_has_prg_ram: bool
    has_trainer: bool
    ignore_mirroring_control: bool
    trainer: bytes | None
    prg_rom: bytes
    chr_rom: bytes


class ApuRegister(IntEnum):
    PULSE1_CONFIG = 0x00
    PULSE1_SWEEP = 0x01
    PULSE1_TIMER = 0x02
    PULSE1_LENGTH_COUNTER = 0x03
    PULSE2_CONFIG = 0x04
    PULSE2_SWEEP = 0x05
    PULSE2_TIMER = 0x06
    PULSE2_LENGTH_COUNTER = 0x07
    TRIANGLE_CONFIG = 0x08
    TRIANGLE_UNUSED = 0x09
    TRIANGLE_TIMER = 0x0A
    TRIANGLE_LENGTH_COUNTER = 0x0B
    NOISE_CONFIG = 0x0C
    NOISE_UNUSED = 0x0D
    NOISE_PERIOD = 0x0E
    NOISE_LENGTH_COUNTER = 0x0F
    DMC_CONFIG = 0x10
    DMC_LOAD_COUNTER = 0x11
    DMC_SAMPLE_ADDRESS = 0x12
    DMC_SAMPLE_LENGTH = 0x13
    STATUS = 0x15
    FRAME_COUNTER = 0x17


APU_LENGTH_TABLE = [
    10,
    254,
    20,
    2,
    40,
    4,
    80,
    6,
    160,
    8,
    60,
    10,
    14,
    12,
    26,
    14,
    12,
    16,
    24,
    18,
    48,
    20,
    96,
    22,
    192,
    24,
    72,
    26,
    16,
    28,
    32,
    30,
]
assert len(APU_LENGTH_TABLE) == 32


@dataclass(eq=False)
class Pulse:
    id: int
    enabled: bool = False
    duty: int = 0
    length_counter_halt: bool = False
    constant_volume: bool = False
    volume: int = 0
    sweep_enabled: bool = False
    sweep_period: int = 0
    sweep_negate: bool = False
    sweep_shift_count: int = 0
    load_timer: int = 0
    load_length_counter: int = 0

    # Internal state
    start_flag: int = 0
    current_tick: int = 0
    current_timer: int = 0
    sweep_divider: int = 0
    length_counter: int = 0
    divider_period: int = 0
    current_sequencer: int = 0
    sweep_reload_flag: int = 0
    decay_level_counter: int = 0
    current_timer_period: int = 0

    def set_enabled(self, value: bool) -> None:
        self.enabled = value
        if not value:
            self.length_counter = 0

    def write_register(self, register: ApuRegister, value: int) -> None:
        if (register & ~0x04) == register.PULSE1_CONFIG:
            self.duty = value >> 6
            self.length_counter_halt = bool(value & 0x20)
            self.constant_volume = bool(value & 0x10)
            self.volume = value & 0xF
            return
        if (register & ~0x04) == register.PULSE1_SWEEP:
            self.sweep_enabled = bool(value & 0x80)
            self.sweep_period = ((value >> 4) & 0x07) + 1
            self.sweep_negate = bool(value & 0x08)
            self.sweep_shift_count = value & 0x07
            # Side effects
            self.sweep_reload_flag = 1
            return
        if (register & ~0x04) == register.PULSE1_TIMER:
            self.load_timer &= ~0xFF
            self.load_timer |= value
            return
        if (register & ~0x04) == register.PULSE1_LENGTH_COUNTER:
            self.load_timer &= ~0x700
            self.load_timer |= (value & 0x7) << 8
            self.load_length_counter = value >> 3
            # Side effects
            if self.enabled:
                self.length_counter = APU_LENGTH_TABLE[self.load_length_counter]
            self.current_timer = self.load_timer
            self.current_sequencer = 0
            self.start_flag = 1
            self.current_timer_period = self.load_timer
            return
        assert False

    def generate(self) -> npt.NDArray[np.uint8]:
        result = np.zeros(Apu.TICKS_IN_FRAME, dtype=np.uint8)
        if not self.enabled:
            return result
        nesapu.generate_pulse(self, result)
        return result


@dataclass(eq=False)
class Triangle:
    enabled: bool = False
    length_counter_halt: bool = False
    load_timer: int = 0
    load_counter: int = 0
    load_length_counter: int = 0

    # Internal state
    current_tick: int = 0
    current_value: int = 0
    current_timer: int = 0
    length_counter: int = 0
    current_counter: int = 0
    current_sequencer: int = 0
    counter_reload_flag: int = 0

    def set_enabled(self, value: bool) -> None:
        self.enabled = value
        if not value:
            self.length_counter = 0

    def write_register(self, register: ApuRegister, value: int) -> None:
        if register == register.TRIANGLE_CONFIG:
            self.length_counter_halt = bool(value & 0x80)
            self.load_counter = value & 0x7F
            return
        if register == register.TRIANGLE_UNUSED:
            return
        if register == register.TRIANGLE_TIMER:
            self.load_timer &= ~0xFF
            self.load_timer |= value
            return
        if register == register.TRIANGLE_LENGTH_COUNTER:
            self.load_timer &= ~0x700
            self.load_timer |= (value & 0x7) << 8
            self.load_length_counter = value >> 3
            # Set length counter
            if self.enabled:
                self.length_counter = APU_LENGTH_TABLE[self.load_length_counter]
            # Reset internal state
            self.counter_reload_flag = 1
            return
        assert False

    def generate(self) -> npt.NDArray[np.uint8]:
        result = np.zeros(Apu.TICKS_IN_FRAME, dtype=np.uint8)
        nesapu.generate_triangle(self, result)
        return result


@dataclass(eq=False)
class Noise:
    enabled: bool = False

    length_counter_halt: bool = False
    constant_volume: bool = False
    volume: int = 0
    noise_mode: bool = False
    noise_period: int = 0
    load_length_counter: int = 0

    # Internal state
    start_flag: int = 0
    current_tick: int = 0
    current_timer: int = 0
    length_counter: int = 0
    shift_register: int = 1
    divider_period: int = 0
    decay_level_counter: int = 0

    PERIOD_TABLE = [
        4,
        8,
        16,
        32,
        64,
        96,
        128,
        160,
        202,
        254,
        380,
        508,
        762,
        1016,
        2034,
        4068,
    ]

    def set_enabled(self, value: bool) -> None:
        self.enabled = value
        if not value:
            self.length_counter = 0

    def write_register(self, register: ApuRegister, value: int) -> None:
        if register == register.NOISE_CONFIG:
            self.length_counter_halt = bool(value & 0x20)
            self.constant_volume = bool(value & 0x10)
            self.volume = value & 0xF
            return
        if register == register.NOISE_UNUSED:
            raise NotImplementedError
            return
        if register == register.NOISE_PERIOD:
            self.noise_mode = bool(value & 0x80)
            self.noise_period = self.PERIOD_TABLE[value & 0xF]
            return
        if register == register.NOISE_LENGTH_COUNTER:
            self.load_length_counter = value >> 3
            # Set length counter
            if self.enabled:
                self.length_counter = APU_LENGTH_TABLE[self.load_length_counter]
            # Reset internal state
            self.start_flag = 1
            return
        assert False

    def generate(self) -> npt.NDArray[np.uint8]:
        result = np.zeros(Apu.TICKS_IN_FRAME, dtype=np.uint8)
        nesapu.generate_noise(self, result)
        return result


@dataclass(eq=False)
class Apu:
    frame_counter_mode: int = 0

    pulse1: Pulse = field(default_factory=lambda: Pulse(1))
    pulse2: Pulse = field(default_factory=lambda: Pulse(2))
    triangle: Triangle = field(default_factory=Triangle)
    noise: Noise = field(default_factory=Noise)
    dmc_enabled: bool = False

    # Filter configuration
    filter1_enabled: bool = True
    filter1_cutoff: float = 90.0
    filter2_enabled: bool = True
    filter2_cutoff: float = 442.0
    filter3_enabled: bool = True
    filter3_cutoff: float = 14000.0

    # Filter values
    filter1_previous_in: float = 0
    filter1_previous_out: float = 0
    filter2_previous_in: float = 0
    filter2_previous_out: float = 0
    filter3_previous_in: float = 0
    filter3_previous_out: float = 0

    TICKS_IN_FRAME = 14890

    def write_register(self, cpu: Cpu, register: int, value: int) -> None:
        register = ApuRegister(register)
        if register == register.FRAME_COUNTER:
            self.frame_counter_mode = value >> 7
            if self.frame_counter_mode & 0x40:
                # Clear frame interrupt flag
                raise NotImplementedError
            return
        if register == register.STATUS:
            self.dmc_enabled = bool(value & 0x10)
            self.noise.set_enabled(bool(value & 0x08))
            self.triangle.set_enabled(bool(value & 0x04))
            self.pulse2.set_enabled(bool(value & 0x02))
            self.pulse1.set_enabled(bool(value & 0x01))
            return
        if register in (
            register.PULSE1_CONFIG,
            register.PULSE1_SWEEP,
            register.PULSE1_TIMER,
            register.PULSE1_LENGTH_COUNTER,
        ):
            self.pulse1.write_register(register, value)
            return
        if register in (
            register.PULSE2_CONFIG,
            register.PULSE2_SWEEP,
            register.PULSE2_TIMER,
            register.PULSE2_LENGTH_COUNTER,
        ):
            self.pulse2.write_register(register, value)
            return
        if register in (
            register.TRIANGLE_CONFIG,
            register.TRIANGLE_UNUSED,
            register.TRIANGLE_TIMER,
            register.TRIANGLE_LENGTH_COUNTER,
        ):
            self.triangle.write_register(register, value)
            return
        if register in (
            register.NOISE_CONFIG,
            register.NOISE_UNUSED,
            register.NOISE_PERIOD,
            register.NOISE_LENGTH_COUNTER,
        ):
            self.noise.write_register(register, value)
            return
        if register in (
            register.DMC_CONFIG,
            register.DMC_LOAD_COUNTER,
            register.DMC_SAMPLE_ADDRESS,
            register.DMC_SAMPLE_LENGTH,
        ):
            if register == register.DMC_LOAD_COUNTER:
                return
            raise NotImplementedError(register)
        assert False

    def generate_dmc(self) -> npt.NDArray[np.uint8]:
        result = np.zeros(Apu.TICKS_IN_FRAME, dtype=np.uint8)
        if not self.dmc_enabled:
            return result
        raise NotImplementedError

    def generate(self, audio: npt.NDArray[np.int16]) -> None:
        # self.pulse1.set_enabled(False)
        # self.pulse2.set_enabled(False)
        # self.triangle.set_enabled(False)
        # self.noise.set_enabled(False)

        pulse1 = self.pulse1.generate()
        pulse2 = self.pulse2.generate()
        triangle = self.triangle.generate()
        noise = self.noise.generate()
        dmc = self.generate_dmc()
        nesapu.apu_mixer(self, pulse1, pulse2, triangle, noise, dmc, audio)


class PpuRegister(IntEnum):
    PPUCTRL = 0
    PPUMASK = 1
    PPUSTATUS = 2
    OAMADDR = 3
    OAMDATA = 4
    PPUSCROLL = 5
    PPUADDR = 6
    PPUDATA = 7


@dataclass(eq=False)
class Ppu:
    cartridge: Cartridge
    oam: bytearray = field(default_factory=lambda: bytearray(256))
    ram: bytearray = field(default_factory=lambda: bytearray(8 * 256))
    palette: bytearray = field(default_factory=lambda: bytearray(32))

    # Registers
    ctrl: int = 0
    mask: int = 0
    status: int = 0

    x_scroll: int = 0
    y_scroll: int = 0
    scroll_toggle: int = 0

    oam_addr: int = 0
    ppu_addr: int = 0
    ppu_addr_toggle: int = 0
    delayed_read: int = 0

    vblank: bool = True
    sprite_zero_hit: bool = False
    x_scroll_before_sprite_zero_hit: int = 0
    y_scroll_before_sprite_zero_hit: int = 0

    # Tracking
    instruction_count_at_last_ppu_status_read: int = 0

    # Changes
    background_pattern_table_address_changed: bool = False
    background_tile_changed: set[tuple[int, int]] = field(default_factory=set)
    background_tiles: np.ndarray = field(
        default_factory=lambda: np.zeros((240 * 2, 256 * 2), dtype=np.uint32)
    )
    background_tiles_with_palette: list[set[tuple[int, int]]] = field(
        default_factory=lambda: [set(), set(), set(), set()]
    )

    # Properties from PPUCTRL

    @property
    def background_pattern_table_address(self) -> int:
        return 0x1000 if (self.ctrl & 0x10) else 0x0000

    @property
    def sprite_pattern_table_address(self) -> int:
        return 0x1000 if (self.ctrl & 0x08) else 0x0000

    @property
    def sprite_size(self) -> tuple[int, int]:
        return (8, 16) if (self.ctrl & 0x20) else (8, 8)

    @property
    def ram_address_increment(self) -> int:
        return 32 if (self.ctrl & 0x04) else 1

    # Properties from PPUMASK

    @property
    def show_background(self) -> bool:
        return bool(self.mask & 0x08)

    @property
    def show_sprites(self) -> bool:
        return bool(self.mask & 0x10)

    def new_vblank(self) -> None:
        self.x_scroll = 0
        self.y_scroll = 0
        self.scroll_toggle = 0
        self.oam_addr = 0
        self.ppu_addr = 0
        self.ppu_addr_toggle = 0
        self.vblank = True
        self.sprite_zero_hit = False
        self.x_scroll_before_sprite_zero_hit = 0
        self.y_scroll_before_sprite_zero_hit = 0
        self.instruction_count_at_last_ppu_status_read = 0

        self.background_tile_changed.clear()
        self.background_palette_changed = False
        self.background_pattern_table_address_changed = False

    def read_register(self, cpu: Cpu, reg: int) -> int:
        if reg == PpuRegister.PPUCTRL:
            return self.ctrl
        if reg == PpuRegister.PPUMASK:
            return self.mask
        if reg == PpuRegister.PPUSTATUS:
            # Clear
            self.ppu_addr = 0
            self.scroll_toggle = 0
            # Tight loop detected
            if (
                cpu.instruction_count
                <= self.instruction_count_at_last_ppu_status_read + 3
            ):
                if not self.sprite_zero_hit:
                    self.x_scroll_before_sprite_zero_hit = self.x_scroll | (
                        (self.ctrl & 0x01) << 8
                    )
                    self.y_scroll_before_sprite_zero_hit = self.y_scroll | (
                        (self.ctrl & 0x02) << 7
                    )
                    self.sprite_zero_hit = True
                else:
                    self.sprite_zero_hit = False
                    self.vblank = True
            self.instruction_count_at_last_ppu_status_read = cpu.instruction_count
            # First read after VBlank
            if self.vblank:
                self.vblank = False
                return 0x80
            # Sprite 0 Hit has not been reached
            if not self.sprite_zero_hit:
                return 0x00
            # Sprite 0 Hit has been reached
            return 0x40
        if reg == PpuRegister.OAMADDR:
            raise NotImplementedError
        if reg == PpuRegister.OAMDATA:
            raise NotImplementedError
        if reg == PpuRegister.PPUSCROLL:
            raise NotImplementedError
        if reg == PpuRegister.PPUADDR:
            raise NotImplementedError
        if reg == PpuRegister.PPUDATA:
            result = self.ppu_read(self.ppu_addr)
            self.ppu_addr += self.ram_address_increment
            return result
        assert False

    def write_register(self, cpu: "Cpu", reg: int, value: int) -> None:
        if reg == PpuRegister.PPUCTRL:
            old_address = self.background_pattern_table_address
            self.ctrl = value
            new_address = self.background_pattern_table_address
            if old_address != new_address:
                self.background_pattern_table_address_changed = True
            return
        if reg == PpuRegister.PPUMASK:
            self.mask = value
            return
        if reg == PpuRegister.PPUSTATUS:
            raise NotImplementedError
        if reg == PpuRegister.OAMADDR:
            self.oam_addr = value
            return
        if reg == PpuRegister.OAMDATA:
            self.oam[self.oam_addr] = value
            return
        if reg == PpuRegister.PPUSCROLL:
            if self.scroll_toggle == 0:
                self.x_scroll = value
            else:
                self.y_scroll = value
            self.scroll_toggle ^= 1
            return
        if reg == PpuRegister.PPUADDR:
            if self.ppu_addr_toggle == 0:
                self.ppu_addr = value << 8
            else:
                self.ppu_addr |= value
            self.ppu_addr_toggle ^= 1
            return
        if reg == PpuRegister.PPUDATA:
            self.ppu_write(self.ppu_addr, value)
            self.ppu_addr += self.ram_address_increment
            return
        assert False, reg

    def write_oam(self, data: bytes) -> None:
        assert len(data) == 256
        self.oam[:] = data

    def ppu_read(self, addr: int) -> int:
        # CHR rom access
        if 0x0000 <= addr < 0x2000:
            result, self.delayed_read = self.delayed_read, self.cartridge.chr_rom[addr]
            return result
        # Ram access
        if 0x2000 <= addr < 0x3000:
            raise NotImplementedError
        # Palette access
        if 0x3F00 <= addr < 0x3F10:
            raise NotImplementedError
        raise ValueError(f"Invalid PPU read: 0x{addr:04x}")

    def ppu_write(self, addr: int, value: int) -> None:
        # Ram access
        if 0x2000 <= addr < 0x3000:
            a_addr = addr & 0x3FF
            b_addr = (addr & 0x3FF) + 0x400
            if 0x2000 <= addr < 0x2400:
                addr = a_addr
            elif 0x2400 <= addr < 0x2800:
                addr = a_addr if self.cartridge.mirroring == "H" else b_addr
            elif 0x2800 <= addr < 0x2C00:
                addr = b_addr if self.cartridge.mirroring == "H" else a_addr
            elif 0x2C00 <= addr < 0x3000:
                addr = b_addr
            else:
                assert False
            if self.ram[addr] != value:
                self.background_tile_changed.update(self.addr_to_indexes(addr))
            self.ram[addr] = value
            return
        # Palette access
        if 0x3F00 <= addr < 0x3F20:
            addr &= 0x1F
            if addr in (0x00, 0x04, 0x08, 0x0C):
                self.palette[addr | 0x10] = value
            elif addr in (0x10, 0x14, 0x18, 0x1C):
                self.palette[addr & ~0x10] = value
            elif addr < 0x10 and self.palette[addr] != value:
                self.background_tile_changed.update(
                    self.background_tiles_with_palette[addr >> 2]
                )
            self.palette[addr] = value
            return
        raise ValueError(f"Invalid PPU write: 0x{addr:04x}")

    def render(self, video: npt.NDArray[np.uint32]) -> None:
        self.render_background_color(video)
        self.render_sprite(video, behind=True)
        self.render_background(video)
        self.render_sprite(video, behind=False)

    def render_background_color(self, video: npt.NDArray[np.uint32]) -> None:
        background_color = self.palette[0]
        video.fill(nesppu.get_color(background_color))

    def update_tiles(self) -> None:
        base_pattern_address = self.background_pattern_table_address
        # Draw everything
        if (
            self.background_pattern_table_address_changed
            or self.background_palette_changed
        ):
            for x_index in range(64):
                for y_index in range(30):
                    self.update_tile(y_index, x_index, base_pattern_address)
            return
        # Draw changes
        for (y_index, x_index) in self.background_tile_changed:
            self.update_tile(y_index, x_index, base_pattern_address)

    def index_to_addr(self, y: int, x: int) -> tuple[int, int]:
        nametable = ((y & 0x20) << 6) | ((x & 0x20) << 5)
        pattern = nametable | ((y & 0b00011111) << 5) | (x & 0b00011111)
        palette = nametable | 0x03C0
        palette |= (y & 0b00011100) << 1
        palette |= (x & 0b00011100) >> 2
        return pattern, palette

    def addr_to_indexes(self, addr: int) -> list[tuple[int, int]]:
        y = ((addr >> 11) & 0x01) << 5
        x = ((addr >> 10) & 0x01) << 5
        addr &= 0x3FF
        if addr < 0x03C0:
            y |= (addr >> 5) & 0x1F
            x |= (addr >> 0) & 0x1F
            return [(y, x)]
        y |= (addr & 0b00111000) >> 1
        x |= (addr & 0b00000111) << 2
        return [(y | dy, x | dx) for dy in range(4) for dx in range(4)]

    def update_tile(
        self, y_index: int, x_index: int, base_pattern_address: int
    ) -> None:
        # Filter
        if y_index in (30, 31, 62, 63):
            return
        # Get nametable
        pattern_ram_address, palette_ram_address = self.index_to_addr(y_index, x_index)
        # Get pattern address
        pattern_address = self.ram[pattern_ram_address]
        pattern_address = (pattern_address << 4) | base_pattern_address
        # Get colors
        palette_address = self.ram[palette_ram_address]
        shift = ((y_index & 0x2) << 1) | (x_index & 0x02)
        palette_address = ((palette_address >> shift) & 0x3) << 2
        colors = self.palette[palette_address + 1 : palette_address + 4]
        # Update palette info
        palette_address >>= 2
        entry = (y_index, x_index)
        if entry not in self.background_tiles_with_palette[palette_address]:
            for tile_set in self.background_tiles_with_palette:
                tile_set.discard(entry)
            self.background_tiles_with_palette[palette_address].add(entry)
        # Get tile
        tile = self.render_tile(pattern_address, bytes(colors))
        # Blit tile
        y_pixel = y_index << 3
        x_pixel = x_index << 3
        if y_index >= 32:
            y_pixel -= 16
        self.background_tiles[y_pixel : y_pixel + 8, x_pixel : x_pixel + 8] = tile

    def render_background(self, video: npt.NDArray[np.uint32]) -> None:
        self.update_tiles()
        if not self.show_background:
            return
        first_row = 8  # Hide first and last row like most monitors
        sprite_zero_hit_y = self.oam[0] + 8
        x_scroll = self.x_scroll | ((self.ctrl & 0x01) << 8)
        nesppu.blit(
            self.background_tiles[:sprite_zero_hit_y, :], video, (-first_row, 0)
        )
        nesppu.blit(
            self.background_tiles[sprite_zero_hit_y:, :],
            video,
            (sprite_zero_hit_y - first_row, 0 - x_scroll),
        )
        nesppu.blit(
            self.background_tiles[sprite_zero_hit_y:, :],
            video,
            (sprite_zero_hit_y - first_row, 512 - x_scroll),
        )

    def render_sprite(
        self, video: npt.NDArray[np.uint32], behind: bool = False
    ) -> None:
        if not self.show_sprites:
            return
        assert self.sprite_size == (8, 8)
        palette = self.palette
        pattern_table_address = self.sprite_pattern_table_address
        first_row = 8  # Hide first and last row like most monitors
        for i in reversed(range(64)):
            y, index, attr, x = self.oam[i * 4 : i * 4 + 4]
            # Below the screen
            if y >= 240:
                continue
            # Behind or in front of the background
            if behind != bool(attr & 0x20):
                continue
            # Pattern address
            pattern_addr = (index << 4) | pattern_table_address
            # Colors
            color_index = attr & 0x03
            palette_addr = 0x10 | (color_index << 2)
            colors = palette[palette_addr + 1 : palette_addr + 4]
            # Tile
            tile = self.render_tile(pattern_addr, bytes(colors))
            # Vertical flip
            if attr & 0x80:
                tile = tile[::-1, :]
            # Horizontal flip
            if attr & 0x40:
                tile = tile[:, ::-1]
            # Blit
            nesppu.blit(tile, video, (y - first_row, x))

    @lru_cache(maxsize=None)
    def render_tile(self, pattern_addr: int, colors: bytes) -> npt.NDArray[np.uint32]:
        result = np.zeros((8, 8), dtype=np.uint32)
        nesppu.render_tile(self.cartridge.chr_rom, pattern_addr, colors, result)
        return result


@dataclass(eq=False)
class Cpu:
    cartridge: Cartridge
    ppu: Ppu
    apu: Apu
    ram: bytearray = field(default_factory=lambda: bytearray(8 * 256))

    # Internal registers
    pc: int = 0
    sp: int = 0
    a: int = 0
    x: int = 0
    y: int = 0

    # Flags
    n: bool = False
    z: bool = False
    c: bool = False
    v: bool = False
    i: bool = False
    d: bool = False

    # Tracking
    frame: int = 0
    instruction_count: int = 0

    # IO
    input_value: int = 0

    @property
    def rom(self) -> bytes:
        return self.cartridge.prg_rom

    # CPU Bus access

    def cpu_read(self, addr: int) -> int:
        # Ram access
        if 0 <= addr < 0x0800:
            return self.ram[addr]
        # Mirror ram access
        if 0x0800 <= addr < 0x2000:
            return self.ram[addr & 0x07FF]
        # Rom access
        if 0x8000 <= addr < 0x10000:
            return self.cartridge.prg_rom[addr - 0x8000]
        # PPU access
        if 0x2000 <= addr < 0x2008:
            return self.ppu.read_register(self, addr & 0x7)
        # APU/IO access
        if 0x4000 <= addr < 0x4014:
            raise NotImplementedError
        # Joystick 1 data
        if addr == 0x4016:
            result = self.input_value & 0x01
            self.input_value >>= 1
            return result
        # Joystick 2 data
        if addr == 0x4017:
            return 0
        # Invalid access
        raise ValueError(f"Invalid read access: 0x{addr:04x} (pc=0x{self.pc:04x})")

    def cpu_write(self, addr: int, value: int) -> None:
        # Ram access
        if 0 <= addr < 0x0800:
            self.ram[addr] = value
            return
        # PPU access
        if 0x2000 <= addr < 0x2008:
            return self.ppu.write_register(self, addr & 0x7, value)
        # APU access
        if 0x4000 <= addr < 0x4014:
            self.apu.write_register(self, addr & 0x1F, value)
            return
        # OAM DMA
        if addr == 0x4014:
            start = value << 8
            stop = (value + 1) << 8
            data = self.ram[start:stop]
            return self.ppu.write_oam(data)
        # Sound channel control
        if addr == 0x4015:
            self.apu.write_register(self, addr & 0x1F, value)
            return
        # Joystick 1 data
        if addr == 0x4016:
            return
        # Joystick 2 data
        if addr == 0x4017:
            self.apu.write_register(self, addr & 0x1F, value)
            return
        # Rom access
        raise ValueError(f"Invalid write access: 0x{addr:04x} (pc=0x{self.pc:04x})")

    # Entry points

    def load_nmi_entrypoint(self) -> None:
        self.frame += 1
        self.pc = self.cpu_read(0xFFFA)
        self.pc |= self.cpu_read(0xFFFB) << 8

    def load_rst_entrypoint(self) -> None:
        self.pc = self.cpu_read(0xFFFC)
        self.pc |= self.cpu_read(0xFFFD) << 8

    # Run CPU instructions

    def run_instructions(self) -> None:
        jmp = 0x4C
        rti = 0x40
        opc = nescpu.run(self)
        # Slow run
        if opc == jmp:
            raise InfiniteLoop()
        assert opc == rti
        return


def parse_ines(source: str) -> Cartridge:

    with open(source, "rb") as input_file:
        header = input_file.read(16)
        assert header[:4] == b"NES\x1a"
        prg_rom_size = header[4] * 16 * 1024
        chr_rom_size = header[5] * 8 * 1024
        flag6, flag7, flag8, flag9, flag10 = header[6:11]

        mapper = (flag6 >> 4) | (flag7 & 0xF0)
        mirroring = "V" if bool(flag6 & 0x01) else "H"
        cartridge_has_prg_ram = bool(flag6 & 0x02)
        has_trainer = bool(flag6 & 0x04)
        ignore_mirroring_control = bool(flag6 & 0x08)

        trainer = input_file.read(512) if has_trainer else None

        prg_rom = input_file.read(prg_rom_size)
        chr_rom = input_file.read(chr_rom_size)

        assert input_file.read() == b""

    return Cartridge(
        mapper,
        mirroring,
        cartridge_has_prg_ram,
        has_trainer,
        ignore_mirroring_control,
        trainer,
        prg_rom,
        chr_rom,
    )


class Nes(Console):
    WIDTH = 256
    HEIGHT = 240 - 16
    FPS = 60
    TICKS_IN_FRAME = 29780

    INPUT_MAP = {
        Console.Input.A: 0x01,
        Console.Input.B: 0x02,
        Console.Input.SELECT: 0x04,
        Console.Input.START: 0x08,
        Console.Input.UP: 0x10,
        Console.Input.DOWN: 0x20,
        Console.Input.LEFT: 0x40,
        Console.Input.RIGHT: 0x80,
    }

    @classmethod
    def add_console_arguments(cls, parser: ArgumentParser) -> None:
        pass

    def __init__(self, parser_args: Namespace) -> None:
        self.current_state = 0
        self.romfile = parser_args.romfile
        self.cartridge = parse_ines(self.romfile)
        self.cpu = Cpu(
            self.cartridge,
            Ppu(self.cartridge),
            Apu(),
        )
        # Run RST
        self.cpu.load_rst_entrypoint()
        try:
            self.cpu.run_instructions()
        except InfiniteLoop:
            pass

    @property
    def apu(self) -> Apu:
        return self.cpu.apu

    @property
    def ppu(self) -> Ppu:
        return self.cpu.ppu

    def set_input(self, input_set: set[Console.Input]) -> None:
        value = sum(self.INPUT_MAP.get(key, 0) for key in input_set)
        self.cpu.input_value = value

    def advance_one_frame(
        self, video: npt.NDArray[np.uint32], audio: npt.NDArray[np.int16]
    ) -> tuple[bool, int]:
        self.ppu.new_vblank()
        self.cpu.load_nmi_entrypoint()
        self.cpu.run_instructions()
        self.ppu.render(video)
        self.apu.generate(audio)
        return True, self.TICKS_IN_FRAME

    def set_current_state(self, state: int) -> None:
        self.current_state = state % 10

    def get_current_state(self) -> int:
        return self.current_state

    def load_state(self) -> None:
        path = f"{self.romfile}.{self.current_state}.state"
        try:
            with open(path, "rb") as f:
                cpu: Cpu = pickle.loads(zlib.decompress(f.read()))
        except OSError:
            return
        cpu.cartridge = self.cartridge
        cpu.ppu.cartridge = self.cartridge
        self.cpu = cpu

    def save_state(self) -> None:
        path = f"{self.romfile}.{self.current_state}.state"
        cpu = deepcopy(self.cpu)
        del cpu.cartridge
        del cpu.ppu.cartridge
        with open(path, "wb") as f:
            f.write(zlib.compress(pickle.dumps(cpu)))


def main(parser_args: tuple[str, ...] | None = None) -> None:
    gambaterm_main(parser_args, console_cls=Nes)


if __name__ == "__main__":
    main()



# cython: language_level=3

cimport numpy as np

cdef unsigned int[64] COLORMAP = [
    0x545454,
    0x001E74,
    0x081090,
    0x300088,
    0x440064,
    0x5C0030,
    0x540400,
    0x3C1800,
    0x202A00,
    0x083A00,
    0x004000,
    0x003C00,
    0x00323C,
    0x000000,
    0x000000,
    0x000000,
    0x989698,
    0x084CC4,
    0x3032EC,
    0x5C1EE4,
    0x8814B0,
    0xA01464,
    0x982220,
    0x783C00,
    0x545A00,
    0x287200,
    0x087C00,
    0x007628,
    0x006678,
    0x000000,
    0x000000,
    0x000000,
    0xECEEEC,
    0x4C9AEC,
    0x787CEC,
    0xB062EC,
    0xE454EC,
    0xEC58B4,
    0xEC6A64,
    0xD48820,
    0xA0AA00,
    0x74C400,
    0x4CD020,
    0x38CC6C,
    0x38B4CC,
    0x3C3C3C,
    0x000000,
    0x000000,
    0xECEEEC,
    0xA8CCEC,
    0xBCBCEC,
    0xD4B2EC,
    0xECAEEC,
    0xECAED4,
    0xECB4B0,
    0xE4C490,
    0xCCD278,
    0xB4DE78,
    0xA8E290,
    0x98E2B4,
    0xA0D6E4,
    0xA0A2A0,
    0x000000,
    0x000000,
]


def get_color(index):
    return (0xFF << 24) | COLORMAP[index]


def blit(
    np.ndarray[np.uint32_t, ndim=2] source,
    np.ndarray[np.uint32_t, ndim=2] destination,
    coordinate,
):
    cdef unsigned int value
    cdef unsigned int source_height = source.shape[0]
    cdef unsigned int source_width = source.shape[1]
    cdef unsigned int destination_height = destination.shape[0]
    cdef unsigned int destination_width = destination.shape[1]
    cdef int y = coordinate[0]
    cdef int x = coordinate[1]

    cdef unsigned int i_source_start = max(0, -y)
    cdef unsigned int j_source_start = max(0, -x)
    cdef unsigned int i_source_stop = max(0, min(<int>source_height, <int>destination_height - y))
    cdef unsigned int j_source_stop = max(0, min(<int>source_width, <int>destination_width - x))
    cdef unsigned int i_source, j_source, i_destination, j_destination

    for i_source in range(i_source_start, i_source_stop):
        i_destination = i_source + y
        for j_source in range(j_source_start, j_source_stop):
            value = source[i_source, j_source]
            if value:
                j_destination = j_source + x
                destination[i_destination, j_destination] = value


def render_tile(char* rom, unsigned short address, char[3] colors, np.ndarray[np.uint32_t, ndim=2] destination):
    cdef unsigned char x, y, lower_value, higher_value, lower_bit, higher_bit, color_index, color
    cdef unsigned int rgb

    for x in range(8):
        lower_value = rom[address + x]
        higher_value = rom[address + x + 8]
        for y in range(8):
            lower_bit = (lower_value >> (7 - y)) & 0x01
            higher_bit = (higher_value >> (7 - y)) & 0x01
            color_index = lower_bit | (higher_bit << 1)
            if color_index == 0:
                destination[x, y] = 0
            else:
                color = colors[color_index - 1]
                destination[x, y] = COLORMAP[color] | <unsigned int>0xff000000


import os
from pathlib import Path
from setuptools import Extension, setup  # type: ignore

# Read the contents of the README file
LONG_DESCRIPTION = (Path(__file__).parent / "README.md").read_text()


# Defer call to `numpy.get_include`
class NumpyIncludePath(os.PathLike):
    def __str__(self) -> str:
        return self.__fspath__()

    def __fspath__(self) -> str:
        import numpy

        include_path: str = numpy.get_include()
        return os.fspath(include_path)


# Extensions
nescpu_extension = Extension(
    "famiterm.nescpu",
    include_dirs=[NumpyIncludePath()],
    sources=["ext/nescpu.pyx"],
)
nesppu_extension = Extension(
    "famiterm.nesppu",
    include_dirs=[NumpyIncludePath()],
    sources=["ext/nesppu.pyx"],
)
nesapu_extension = Extension(
    "famiterm.nesapu",
    include_dirs=[NumpyIncludePath()],
    sources=["ext/nesapu.pyx"],
)


setup(
    name="famiterm",
    version="0.1.1",
    packages=["famiterm"],
    setup_requires=["setuptools>=42", "Cython>=0.29.13", "numpy"],
    ext_modules=[
        nescpu_extension,
        nesppu_extension,
        nesapu_extension,
    ],
    install_requires=["numpy>=1.20", "gambaterm>=0.12.0"],
    extras_require={
        "controller-support": ["pygame>=1.9.5"],
    },
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "famiterm = famiterm:main",
            "famiterm-ssh = famiterm.ssh:main",
        ],
    },
    package_data={"famiterm": ["py.typed"]},
    description="A NES emulator running in the terminal",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/vxgmichel/famiterm",
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    author="Vincent Michel",
    author_email="vxgmichel@gmail.com",
)



from __future__ import annotations

from .run import Nes
from gambaterm.ssh import main as gambaterm_ssh_main


def main(parser_args: tuple[str, ...] | None = None) -> None:
    gambaterm_ssh_main(parser_args, console_cls=Nes)
